#!/bin/awk -f
#!/opt/busybox/awk -f

# EXPERIMENTAL Markdown processor with minimal dependencies.
# Meant to support all features of John Grubers basic Markdown
# + a number of common extensions, mostly inspired by Pandoc Markdown

# Copyright 2021 - 2023 Paul Hänsch
# 
# Permission to use, copy, modify, and/or distribute this software for any
# purpose with or without fee is hereby granted, provided that the above
# copyright notice and this permission notice appear in all copies.
# 
# THE SOFTWARE IS PROVIDED “AS IS” AND THE AUTHOR DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY
# SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR
# IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

# Supported Features / TODO:
# ==========================
# [x] done    [ ] todo    [-] not planned    ? unsure
#
# Basic Markdown - Block elements:
# -------------------------------
# - [x] Paragraphs
#   - [x] Double space line breaks
# - [x] Proper block element nesting
# - [x] Headings
# - [x] ATX-Style Headings
# - [x] Blockquotes
# - [x] Lists (ordered, unordered)
# - [x] Code blocks (using indention)
# - [x] Horizontal rules
# - [x] Verbatim HTML block (disabled by default)
#
# Basic Markdown - Inline elements:
# ---------------------------------
# - [x] Links
# - [x] Reference style links
# - [x] Emphasis *em*/**strong** (*Asterisk*, _Underscore_)
# - [x] `code`, also ``code containing `backticks` ``
# - [x] Images / reference style images
# - [x] <automatic links>
# - [x] backslash escapes
# - [x] Verbatim HTML inline (disabled by default)
# - [x] HTML escaping
#
# NOTE: Set the environment variable MD_HTML=true to enable verbatim HTML
#
# Extensions - Block elements:
# ----------------------------
# - [x] Automatic <section>-wrapping (custom)
# -  ?  Heading identifiers (php md, pandoc)
#   - [x] Heading attributes (custom)
#   - [ ] <hr> terminates section
# - [x] Automatic heading identifiers (custom)
# - [x] Fenced code blocks (php md, pandoc)
#   - [x] Fenced code attributes
# - [x] Images (as block elements, <figure>-wrapped) (custom)
#   - [x] reference style block images
# - [/] Tables
#   -  ?  Simple table (pandoc)
#   -  ?  Multiline table (pandoc)
#   - [x] Grid table (pandoc)
#     - [x] Headerless
#   - [x] Pipe table (php md, pandoc)
# - [x] Line blocks (pandoc)
# - [x] Task lists (pandoc, custom)
# - [x] Definition lists (php md, pandoc)
# - [-] Numbered example lists (pandoc)
# - [-] Metadata blocks (pandoc)
# - [x] Metadata blocks (custom)
# - [x] Fenced Divs (pandoc)
#
# Extensions - Inline elements:
# ----------------------------
# - [x] Ignore embedded_underscores (php md, pandoc)
# - [x] ~~strikeout~~ (pandoc)
# - [x] ^Superscript^ ~Subscript~ (pandoc)
# - [-] Bracketed spans (pandoc)
#   - [-] Inline attributes (pandoc)
# - [x] Image attributes (custom, pandoc inspired, not for reference style)
# - [x] Wiki style links [[PageName]] / [[PageName|Link Text]]
# - [-] TEX-Math (pandoc)
# -  ?  Footnotes (php md)
# -  ?  Abbreviations (php md)
# -  ?  "Curly quotes" (smartypants)
# - [ ] em-dashes (--) (smartypants old)
# -  ?  ... three-dot ellipsis (smartypants)
# - [-] en-dash (smartypants)
# - [ ] Automatic em-dash / en-dash
# - [x] Automatic -> Arrows <- (custom)

function debug(text) { printf "\n---\n%s\n---\n", text > "/dev/stderr"; }

function HTML ( text ) {
  gsub( /&/,  "\\&amp;",  text );
  gsub( /</,  "\\&lt;",   text );
  gsub( />/,  "\\&gt;",   text );
  gsub( /"/,  "\\&quot;", text );
  gsub( /'/,  "\\&#x27;", text );
  gsub( /\\/, "\\&#x5C;", text );
  return text;
}

function URL ( text, sharp ) {
  gsub( /&/,  "%26",  text );
  gsub( /"/,  "%22", text );
  gsub( /'/,  "%27", text );
  gsub( /`/,  "%60", text );
  gsub( /\?/,  "%3F", text );
  if (sharp) gsub( /#/,  "%23", text );
  gsub( /\[/,  "%5B", text );
  gsub( /\]/,  "%5D", text );
  gsub( / /,  "%20", text );
  gsub( /	/,  "%09", text );
  gsub( /\\/, "%5C", text );
  return text;
}

function inline( line, LOCAL, len, text, code, href, guard ) {
  if ( line ~ /^$/ ) {  # Recursion End
    return "";

  # omit processing of escaped characters
  } else if ( line ~ /^\\./) {
    return HTML(substr(line, 2, 1)) inline( substr(line, 3) );

  # hard brakes
  } else if ( match(line, /^  \n/) ) {
    return "<br>\n" inline( substr(line, RLENGTH + 1) );

  #  ``code spans``
  } else if ( match( line, /^`+/) ) {
    len = RLENGTH
    guard = substr( line, 1, len )
    if ( match(line, guard ".*" guard) ) {
      code = substr( line, len + 1, match( substr(line, len + 1), guard ) - 1)
      len = 2 * length(guard) + length(code)
      #  strip single surrounding white spaces
      gsub( /^ | $/, "", code)
      #  escape HTML within code span
      gsub( /&/, "\\&amp;", code ); gsub( /</, "\\&lt;", code ); gsub( />/, "\\&gt;", code );
      return "<code>" code "</code>" inline( substr( line, len + 1 ) )
    }

  # Macros
  } else if ( match( line, /^<<([^>]|>[^>])+>>/ ) ) {
    len = RLENGTH;
    return "<code class=\"macro\">" HTML( substr( line, 3, len - 4 ) ) "</code>" inline(substr(line, len + 1));

  # Wiki style links
  } else if ( match( line, /^\[\[([^]|]+)(\|[^]]+)?\]\]/) ) {
    len = RLENGTH;
    href = gensub(/^\[\[([^]|]+)(\|([^]]+))?\]\]/, "\\1", 1, substr(line, 1, len) );
    text = gensub(/^\[\[([^]|]+)(\|([^]]+))?\]\]/, "\\3", 1, substr(line, 1, len) );
    if ( ! text ) text = href;
    return "<a href=\"" HTML(href) "\">" HTML(text) "</a>" inline( substr( line, len + 1) );

  #  quick links ("automatic links" in md doc)
  } else if ( match( line, /^<[a-zA-Z]+:\/\/([-\.[:alnum:]]+)(:[0-9]*)?(\/[^>]*)?>/ ) ) {
    len = RLENGTH;
    href = HTML( substr( line, 2, len - 2) );
    return "<a href=\"" href "\">" href "</a>" inline( substr( line, len + 1) );

  # quick link email
  } else if ( match( line, /^<[a-zA-Z0-9.!#$%&'\''*+\/=?^_`{|}~-]+@[a-zA-Z0-9]([a-zA-Z0-9-]{0,61}[a-zA-Z0-9])?(\.[a-zA-Z0-9]([a-zA-Z0-9-]{0,61}[a-zA-Z0-9])?)*>/ ) ) {
    len = RLENGTH;
    href = HTML( substr( line, 2, len - 2) );
    return "<a href=\"mailto:" href "\">" href "</a>" inline( substr( line, len + 1) );

  # Verbatim inline HTML
  } else if ( AllowHTML && match( line, /^(<!--([^-]|-[^-]|--[^>])*-->|<\?([^\?]|\?[^>])*\?>|<![A-Z][^>]*>|<!\[CDATA\[([^\]]|\][^\]]|\]\][^>])*\]\]>|<\/[A-Za-z][A-Za-z0-9-]*[[:space:]]*>|<[A-Za-z][A-Za-z0-9-]*([[:space:]]+[A-Za-z_:][A-Za-z0-9_\.:-]*([[:space:]]*=[[:space:]]*([[:space:]"'=<>`]+|"[^"]*"|'[^']*'))?)*[[:space:]]*\/?>)/) ) {
    len = RLENGTH;
    return substr( line, 1, len) inline(substr(line, len + 1));

  # inline links
  } else if ( match(line, "^" lii "\\([\n\t ]*" lid "([\n\t ]+" lit ")?[\n\t ]*\\)") ) {
    len = RLENGTH;
    text = href = title = substr( line, 1, len);
    sub("^\\[", "", text); sub("\\]\\([\n\t ]*" lid "([\n\t ]+" lit ")?[\n\t ]*\\)$", "", text);
    sub("^" lii "\\([\n\t ]*", "", href); sub("([\n\t ]+" lit ")?[\n\t ]*\\)$", "", href);
    sub("^" lii "\\([\n\t ]*" lid, "", title); sub("[\n\t ]*\\)$", "", title); sub("^[\n\t ]+", "", title);

    if ( match(href, /^<.*>$/) ) { sub(/^</, "", href); sub(/>$/, "", href); }
         if ( match(title, /^".*"$/) ) { sub(/^"/, "", title); sub(/"$/, "", title); }
    else if ( match(title, /^'.*'$/) ) { sub(/^'/, "", title); sub(/'$/, "", title); }
    else if ( match(title, /^\(.*\)$/) ) { sub(/^\(/, "", title); sub(/\)$/, "", title); }

    gsub(/\\/, "", href); gsub(/\\/, "", title); gsub(/[\n\t]+/, " ", title);

    return "<a href=\"" HTML(href) "\"" (title?" title=\"" HTML(title) "\"":"") ">" \
           inline( text ) "</a>" inline( substr( line, len + 1) );

  # reference style links
  } else if ( match(line, /^\[([^]]+)\] ?\[([^]]*)\]/ ) ) {
    len = RLENGTH;
    text = gensub(/^\[([^\n]+)\] ?\[([^\n]*)\].*/, "\\1", 1, substr(line, 1, len) );
      id = gensub(/^\[([^\n]+)\] ?\[([^\n]*)\].*/, "\\2", 1, substr(line, 1, len) );
    if ( ! id ) id = text;
    if ( rl_href[id] && rl_title[id] ) {
      return "<a href=\"" HTML(rl_href[id]) "\" title=\"" HTML(rl_title[id]) "\">" inline(text) "</a>" inline( substr( line, len + 1) );
    } else if ( rl_href[id] ) {
      return "<a href=\"" HTML(rl_href[id]) "\">" inline(text) "</a>" inline( substr( line, len + 1) );
    } else {
      return "" HTML(substr(line, 1, len)) inline( substr(line, len + 1) );
    }

  # inline images
  } else if ( match(line, "^!" lix "\\([\n\t ]*" lid "([\n\t ]+" lit ")?[\n\t ]*\\)(\\{[a-zA-Z \t-]*\\})?") ) {
    len = RLENGTH; text = href = title = attrib = substr( line, 1, len);

    sub("^!\\[", "", text);
    sub("\\]\\([\n\t ]*" lid "([\n\t ]+" lit ")?[\n\t ]*\\)(\\{[a-zA-Z \t-]*\\})?$", "", text);

    sub("^!" lix "\\([\n\t ]*", "", href);
    sub("([\n\t ]+" lit ")?[\n\t ]*\\)(\\{[a-zA-Z \t-]*\\})?$", "", href);

    sub("^!" lix "\\([\n\t ]*" lid, "", title);
    sub("[\n\t ]*\\)(\\{[a-zA-Z \t-]*\\})?$", "", title);
    sub("^[\n\t ]+", "", title);

    sub("^!" lix "\\([\n\t ]*" lid "([\n\t ]+" lit ")?[\n\t ]*\\)", "", attrib);
    sub(/^\{[ \t]*/, "", attrib); sub(/[ \t]*\}$/, "", attrib); gsub(/[ \t]+/, " ", attrib);

    if ( match(href, /^<.*>$/) ) { sub(/^</, "", href); sub(/>$/, "", href); }
         if ( match(title, /^".*"$/) ) { sub(/^"/, "", title); sub(/"$/, "", title); }
    else if ( match(title, /^'.*'$/) ) { sub(/^'/, "", title); sub(/'$/, "", title); }
    else if ( match(title, /^\(.*\)$/) ) { sub(/^\(/, "", title); sub(/\)$/, "", title); }

    gsub(/^[\t ]+$/, "", text); gsub(/\\/, "", href);
    gsub(/\\/, "", title); gsub(/[\n\t]+/, " ", title);

    return "<img src=\"" HTML(href, 1) "\" alt=\"" HTML(text?text:title?title:href) "\"" \
           (title?" title=\"" HTML(title) "\"":"") (attrib?" class=\"" HTML(attrib) "\"":"") \
           ">" inline( substr( line, len + 1) );

  # reference style images
  } else if ( match(line, /^!\[([^]]*)\] ?\[([^]]*)\]/ ) ) {
    len = RLENGTH;
    text = gensub(/^!\[([^\n]*)\] ?\[([^\n]*)\].*/, "\\1", 1, substr(line, 1, len) );
      id = gensub(/^!\[([^\n]*)\] ?\[([^\n]*)\].*/, "\\2", 1, substr(line, 1, len) );
    if ( ! id ) id = text;
    if ( rl_href[id] && rl_title[id] ) {
      return "<img src=\"" HTML(rl_href[id], 1) "\" alt=\"" HTML(text) "\" title=\"" HTML(rl_title[id]) "\">" \
             inline( substr( line, len + 1) );
    } else if ( rl_href[id] ) {
      return "<img src=\"" HTML(rl_href[id], 1) "\" alt=\"" HTML(text) "\">" \
             inline( substr( line, len + 1) );
    } else {
      return "" HTML(substr(line, 1, len)) inline( substr(line, len + 1) );
    }

  #  ~~strikeout~~ (pandoc)
  } else if ( match(line, /^~~([[:graph:]]|[[:graph:]]([^~]|~[^~])*[[:graph:]])~~/) ) {
    len = RLENGTH;
    return "<del>" inline( substr( line, 3, len - 4 ) ) "</del>" inline( substr( line, len + 1 ) );

  #  ^superscript^ (pandoc)
  } else if ( match(line, /^\^([^[:space:]^]|\\[ ^])+\^/) ) {
    len = RLENGTH;
    return "<sup>" inline( substr( line, 2, len - 2 ) ) "</sup>" inline( substr( line, len + 1 ) );

  #  ~subscript~ (pandoc)
  } else if ( match(line, /^~([^[:space:]~]|\\[ ~])+~/) ) {
    len = RLENGTH;
    return "<sub>" inline( substr( line, 2, len - 2 ) ) "</sub>" inline( substr( line, len + 1 ) );

  # ignore embedded underscores (pandoc, php md)
  } else if ( match(line, "^[[:alnum:]](__|_)") ) {
    return HTML(substr( line, 1, RLENGTH)) inline( substr(line, RLENGTH + 1) );

  #  __strong__$
  } else if ( match(line, "^__(([^_[:space:]]|" ieu ")|([^_[:space:]]|" ieu ")(" nu "|" ieu ")*([^_[:space:]]|" ieu "))__$") ) {
    len = RLENGTH;
    return "<strong>" inline( substr( line, 3, len - 4 ) ) "</strong>" inline( substr( line, len + 1 ) );

  #  __strong__
  } else if ( match(line, "^__(([^_[:space:]]|" ieu ")|([^_[:space:]]|" ieu ")(" nu "|" ieu ")*([^_[:space:]]|" ieu "))__[[:space:][:punct:]]") ) {
    len = RLENGTH;
    return "<strong>" inline( substr( line, 3, len - 5 ) ) "</strong>" inline( substr( line, len) );

  #  **strong**
  } else if ( match(line, "^\\*\\*(([^\\*[:space:]]|" iea ")|([^\\*[:space:]]|" iea ")(" na "|" iea ")*([^\\*[:space:]]|" iea "))\\*\\*") ) {
    len = RLENGTH;
    return "<strong>" inline( substr( line, 3, len - 4 ) ) "</strong>" inline( substr( line, len + 1 ) );

  #  _em_$
  } else if ( match(line, "^_(([^_[:space:]]|" isu ")|([^_[:space:]]|" isu ")(" nu "|" isu ")*([^_[:space:]]|" isu "))_$") ) {
    len = RLENGTH;
    return "<em>" inline( substr( line, 2, len - 2 ) ) "</em>" inline( substr( line, len + 1 ) );

  #  _em_
  } else if ( match(line, "^_(([^_[:space:]]|" isu ")|([^_[:space:]]|" isu ")(" nu "|" isu ")*([^_[:space:]]|" isu "))_[[:space:][:punct:]]") ) {
    len = RLENGTH;
    return "<em>" inline( substr( line, 2, len - 3 ) ) "</em>" inline( substr( line, len ) );

  #  *em*
  } else if ( match(line, "^\\*(([^\\*[:space:]]|" isa ")|([^\\*[:space:]]|" isa ")(" na "|" isa ")*([^\\*[:space:]]|" isa "))\\*") ) {
    len = RLENGTH;
    return "<em>" inline( substr( line, 2, len - 2 ) ) "</em>" inline( substr( line, len + 1 ) );

  # Literal HTML entities
  } else if ( match( line, /^&([a-zA-Z]{2,32}|#[0-9]{1,7}|#[xX][0-9a-fA-F]{1,6});/) ) {
    len = RLENGTH;
    return substr( line, 1, len ) inline(substr(line, len + 1));

  # Arrows
  } else if ( line ~ /^-->( |$)/) {  # ignore multidash-arrow
    return "--&gt;" inline( substr(line, 4) );
  } else if ( line ~ /^<-( |$)/) {
    return "&larr;" inline( substr(line, 3) );
  } else if ( line ~ /^->( |$)/) {
    return "&rarr;" inline( substr(line, 3) );

  # Escape lone HTML character
  } else if ( match( line, /^[&<>"']/) ) {
    return HTML(substr(line, 1, 1)) inline(substr(line, 2));

  #  continue walk over string
  } else {
    return substr(line, 1, 1) inline( substr(line, 2) );
  }
}

function headline( hlvl, htxt, attrib, LOCAL, sec, n, HL) {
  match(hstack, /([0-9]+( [0-9]+){5})$/); split( substr(hstack, RSTART),  HL);

  for ( n = hlvl; n <= 6; n++ ) { sec = sec (HL[n]?"</section>":""); }
  HL[hlvl]++; for ( n = hlvl + 1; n <= 6; n++) { HL[n] = 0;}

  hid = ""; for ( n = 2; n <= blvl; n++) { hid = hid BL[n] "/"; }
  hid = hid HL[1]; for ( n = 2; n <= hlvl; n++) { hid = hid "." HL[n] ; }
  hid = hid ":" URL(htxt, 1);

  sub(/([0-9]+( [0-9]+){5})$/, "", hstack);
  hstack = hstack HL[1] " " HL[2] " " HL[3] " " HL[4] " " HL[5] " " HL[6];

  return sec "<section class=\"" (attrib ? "h" hlvl " " attrib : "h" hlvl)  "\" id=\"" hid "\">" \
         "<h" hlvl (attrib ? " class=\"" attrib "\"" : "") ">" inline( htxt ) \
         "<a class=\"anchor\" href=\"#" hid "\"></a>" \
         "</h" hlvl ">\n";
}

# Nested Block, resets heading counters
function _nblock( block, LOCAL, sec, n ) {
  hstack = hstack " 0 0 0 0 0 0";

  # Block Level
  blvl++; BL[blvl]++;
  for ( n = blvl + 1; n in BL; n++) { delete BL[n]; }

  block = _block( block );
  match(hstack, /([0-9]+( [0-9]+){5})$/); split( substr(hstack, RSTART),  HL);
  sec = ""; for ( n = 1; n <= 6; n++ ) { sec = sec (HL[n]?"</section>":""); }

  sub("( +[0-9]+){6} *$", "", hstack); blvl--;
  return block sec;
}

function _block( block, LOCAL, st, len, text, title, attrib, href, guard, code, indent, list ) {
  gsub( "(^\n+|\n+$)", "", block );

  if ( block == "" ) {
    return "";

  # HTML #2 #3 #4 $5
  } else if ( AllowHTML && match( block, /(^|\n) ? ? ?(<!--([^-]|-[^-]|--[^>])*(-->|$)|<\?([^\?]|\?[^>])*(\?>|$)|<![A-Z][^>]*(>|$)|<!\[CDATA\[([^\]]|\][^\]]|\]\][^>])*(\]\]>|$))/) ) {
    len = RLENGTH; st = RSTART;
    return _block(substr(block, 1, st - 1)) substr(block, st, len) _block(substr(block, st + len));

  # HTML #6
  } else if ( AllowHTML && match( tolower(block), /(^|\n) ? ? ?<\/?(address|article|aside|base|basefont|blockquote|body|caption|center|col|colgroup|dd|details|dialog|dir|div|dl|dt|fieldset|figcaption|figure|footer|form|frame|frameset|h[123456]|head|header|hr|html|iframe|legend|li|link|main|menu|menuitem|nav|noframes|ol|optgroup|option|p|param|section|source|summary|table|tbody|td|tfoot|th|thead|title|tr|track|ul)([[:space:]\n>]|\/>)([^\n]|\n[ \t]*[^\n])*(\n[[:space:]]*\n|$)/) ) {
    len = RLENGTH; st = RSTART;
    return _block(substr(block, 1, st - 1)) substr(block, st, len) _block(substr(block, st + len));

  # HTML #1
  } else if ( AllowHTML && match( tolower(block), /(^|\n) ? ? ?<(script|pre|style)([[:space:]\n>]).*(<\/script>|<\/pre>|<\/style>|$)/) ) {
    len = RLENGTH; st = RSTART;
    match( tolower(substr(block, st, len)), /(<\/script>|<\/pre>|<\/style>)/);
    len = RSTART + RLENGTH;
    return _block(substr(block, 1, st - 1)) substr(block, st, len) _block(substr(block, st + len));

  # HTML #7
  } else if ( AllowHTML && match( block, /^ ? ? ?(<\/[A-Za-z][A-Za-z0-9-]*[[:space:]]*>|<[A-Za-z][A-Za-z0-9-]*([[:space:]]+[A-Za-z_:][A-Za-z0-9_\.:-]*([[:space:]]*=[[:space:]]*([[:space:]"'=<>`]+|"[^"]*"|'[^']*'))?)*[[:space:]]*\/?>)([[:space:]]*\n)([^\n]|\n[ \t]*[^\n])*(\n[[:space:]]*\n|$)/) ) {
    len = RLENGTH; st = RSTART;
    return substr(block, st, len) _block(substr(block, st + len));

  # Metadata (custom, block starting with %something)
  # Metadata is ignored but can be interpreted externally
  } else if ( match(block, /^%[a-zA-Z-]+([[:space:]][^\n]*)?(\n|$)(%[a-zA-Z-]+([[:space:]][^\n]*)?(\n|$)|%([[:space:]][^\n]*)?(\n|$)|[ \t]+[^\n[:space:]][^\n]*(\n|$))*/) ) {
    len = RLENGTH; st = RSTART;
    return  _block( substr( block, len + 1) );
 
  # Blockquote (leading >)
  } else if ( match( block, /^> /) ) {
    match( block, /(^|\n)[[:space:]]*(\n|$)/ ) || match(block, /$/);
    len = RLENGTH; st = RSTART;
    text = substr(block, 1, st - 1); gsub( /(^|\n)> /, "\n", text );
    text = _nblock( text ); gsub( /^\n|\n$/, "", text )
    return "<blockquote>" text "</blockquote>\n\n" _block( substr(block, st + len) );

  # Pipe Tables (pandoc / php md / gfm )
  } else if ( match(block, "^((\\|)?([^\n]+\\|)+[^\n]+(\\|)?)\n" \
                           "((\\|)?(:?-+:?[\\|+])+:?-+:?(\\|)?)\n" \
                           "((\\|)?([^\n]+\\|)+[^\n]+(\\|)?(\n|$))+" ) ) {
    len = RLENGTH; st = RSTART;
    #initialize empty arrays
    split("", talign); split("", tarray);
    cols = 0; cnt=0; ttext = "";

    # table header and alignment
    split( gensub( /(^\||\|$)/, "", "g", \
           gensub( /(^|[^\\])\\\|/, "\\1\\&#x7C;", "g", \
           substr(block, 1, match(block, /(\n|$)/)) \
    )), tarray, /\|/);
    block = substr(block, match(block, /(\n|$)/) + 1 );
    cols = split( \
           gensub( /(^\||\|$)/, "", "g", \
           substr(block, 1, match(block, /(\n|$)/)) \
    ), talign, /[+\|]/);
    block = substr(block, match(block, /(\n|$)/) + 1 );

    for( cnt = 1; cnt < cols; cnt++ ) {
           if (match(talign[cnt], /:-+:/)) talign[cnt]="center";
      else if (match(talign[cnt],  /-+:/)) talign[cnt]="right";
      else if (match(talign[cnt],  /:-+/)) talign[cnt]="left";
      else talign[cnt]="";
    }

    ttext = "<thead>\n<tr>"
    for (cnt = 1; cnt < cols; cnt++)
      ttext = ttext "<th align=\"" talign[cnt] "\">" inline(tarray[cnt]) "</th>"
    ttext = ttext "</tr>\n</thead><tbody>\n"

    while ( match(block, "^((\\|)?([^\n]+\\|)+[^\n]+(\\|)?(\n|$))+" ) ){
      split( gensub( /(^\||\|$)/, "", "g", \
             gensub( /(^|[^\\])\\\|/, "\\1\\&#x7C;", "g", \
             substr(block, 1, match(block, /(\n|$)/)) \
      )), tarray, /\|/);
      block = substr(block, match(block, /(\n|$)/) + 1 );

      ttext = ttext "<tr>"
      for (cnt = 1; cnt < cols; cnt++)
        ttext = ttext "<td align=\"" talign[cnt] "\">" inline(tarray[cnt]) "</td>"
      ttext = ttext "</tr>\n"
    }
    return "<table>" ttext "</tbody></table>\n" _block(block);

  # Grid Tables (pandoc)
  # (with, and without header)
  } else if ( match( block, "^\\+(-+\\+)+\n" \
                            "(\\|([^\n]+\\|)+\n)+" \
                            "(\\+(:?=+:?\\+)+)\n" \
                           "((\\|([^\n]+\\|)+\n)+" \
                             "\\+(-+\\+)+(\n|$))+" \
                   ) || \
              match( block, "^()()()" \
                            "(\\+(:?-+:?\\+)+)\n" \
                           "((\\|([^\n]+\\|)+\n)+" \
                             "\\+(-+\\+)+(\n|$))+" \
  ) ) {
    len = RLENGTH; st = RSTART;
    #initialize empty arrays
    split("", talign); split("", tarray); split("", tread);
    cols = 0; cnt=0; ttext = "";

    # Column Count
    cols = split(   gensub( "^(\\+(:?-+:?\\+)+)(\n.*)*$", "\\1", 1, block), tread, /\+/) - 2;
    # debug(" Cols: " gensub( "^(\\+(:?-+:?\\+)+)(\n.*)*$", "\\1", 1, block ));

    # table alignment
    split( gensub( "^(.*\n)?\\+((:?=+:?\\+|(:-+|-+:|:-+:)\\+)+)(\n.*)$", "\\2", "g", block ), talign, /\+/ );
    # debug("Align: " gensub( "^(.*\n)?\\+((:?=+:?\\+|(:-+|-+:|:-+:)\\+)+)(\n.*)$", "\\2", "g", block ));

    for (cnt = 1; cnt <= cols; cnt++) {
           if (match(talign[cnt], /:(-+|=+):/)) talign[cnt]="center";
      else if (match(talign[cnt],  /(-+|=+):/)) talign[cnt]="right";
      else if (match(talign[cnt], /:(-+|=+)/ )) talign[cnt]="left";
      else talign[cnt]="";
    }

    if ( match(block, "^\\+(-+\\+)+\n" \
                      "(\\|([^\n]+\\|)+\n)+" \
                       "\\+(:?=+:?\\+)+\n" \
                     "((\\|([^\n]+\\|)+\n)+" \
                       "\\+(-+\\+)+(\n|$))+" \
    ) ) {
      # table header
      block = substr(block, match(block, /(\n|$)/) + 1 );
      while ( match(block, "^\\|([^\n]+\\|)+\n") ) {
        split( gensub( /(^\||\|$)/, "", "g", \
                 gensub( /(^|[^\\])\\\|/, "\\1\\&#x7C;", "g", \
                   substr(block, 1, match(block, /(\n|$)/)) \
        )), tread, /\|/);
        block = substr(block, match(block, /(\n|$)/) + 1 );
        for (cnt = 1; cnt <= cols; cnt++)
          tarray[cnt] = tarray[cnt] "\n" tread[cnt];
      }

      ttext = "<thead>\n<tr>"
      for (cnt = 1; cnt <= cols; cnt++)
        ttext = ttext "<th align=\"" talign[cnt] "\">" _nblock(tarray[cnt]) "</th>"
      ttext = ttext "</tr>\n</thead>"
    }

    # table body
    block = substr(block, match(block, /(\n|$)/) + 1 );
    ttext = ttext "<tbody>\n"

    while ( match(block, /^((\|([^\n]+\|)+\n)+\+(-+\+)+(\n|$))+/ ) ){
      split("", tarray);
      while ( match(block, /^\|([^\n]+\|)+\n/) ) {
        split( gensub( /(^\||\|$)/, "", "g", \
               gensub( /(^|[^\\])\\\|/, "\\1\\&#x7C;", "g", \
               substr(block, 1, match(block, /(\n|$)/)) \
        )), tread, /\|/);
        block = substr(block, match(block, /(\n|$)/) + 1 );
        for (cnt = 1; cnt <= cols; cnt++)
          tarray[cnt] = tarray[cnt] "\n" tread[cnt];
      }
      block = substr(block, match(block, /(\n|$)/) + 1 );

      ttext = ttext "<tr>"
      for (cnt = 1; cnt <= cols; cnt++)
        ttext = ttext "<td align=\"" talign[cnt] "\">" _nblock(tarray[cnt]) "</td>"
      ttext = ttext "</tr>\n"
    }
    return "<table>" ttext "</tbody></table>\n" _nblock(block);

  # Line Blocks (pandoc)
  } else if ( match(block, /^\| [^\n]*(\n|$)(\| [^\n]*(\n|$)|[ \t]+[^\n[:space:]][^\n]*(\n|$))*/) ) {
    len = RLENGTH; st = RSTART;

    text = substr(block, 1, len); gsub(/\n[[:space:]]+/, " ", text);
    gsub(/\n\| /, "\n", text); gsub(/^\| |\n$/, "", text);
    text = inline(text); gsub(/\n/, "<br>\n", text);

    return "<div class=\"line-block\">" text "</div>\n" _block( substr( block, len + 1) );

  # Indented Code Block
  } else if ( match(block, /^(    |\t)( *\t*[^ \t\n]+ *\t*)+(\n|$)((    |\t)[^\n]+(\n|$)|[ \t]*(\n|$))*/) ) {
    len = RLENGTH; st = RSTART;
    code = substr(block, 1, len);
    gsub(/(^|\n)(    |\t)/, "\n", code);
    gsub(/^\n|\n+$/, "", code);
    return "<pre><code>" HTML( code ) "</code></pre>\n" \
           _block( substr( block, len + 1 ) );

  # Fenced Divs (pandoc, custom)
  } else if ( match( block, /^(:::+)/ ) ) {
    guard = substr( block, 1, RLENGTH );
    code = block; sub(/^[^\n]+\n/, "", code);
    attrib = gensub(/^:::+[ \t]*\{?[ \t]*([^\}\n]*)\}?[ \t]*\n.*$/, "\\1", 1, block);
    gsub(/[^a-zA-Z0-9_-]+/, " ", attrib);
    gsub(/(^ | $)/, "", attrib);
    if ( match(code, "(^|\n)" guard "+(\n|$)" ) ) {
      len = RLENGTH; st = RSTART;
      return "<div class=\"" attrib "\">" _nblock( substr(code, 1, st - 1) ) "</div>\n" \
             _block( substr( code, st + len ) );
    } else {
      match( block, /(^|\n)[[:space:]]*(\n|$)/ ) || match( block, /$/ );
      len = RLENGTH; st = RSTART;
      return "<p>" inline( substr(block, 1, st - 1) ) "</p>\n" \
             _block( substr(block, st + len) );
    }

  # Fenced Code Block (pandoc)
  } else if ( match( block, /^(~~~+|```+)/ ) ) {
    guard = substr( block, 1, RLENGTH );
    code = gensub(/^[^\n]+\n/, "", 1, block);
    attrib = gensub(/^(~~~+|```+)[ \t]*\{?[ \t]*([^\}\n]*)\}?[ \t]*\n.*$/, "\\2", 1, block);
    gsub(/[^a-zA-Z0-9_-]+/, " ", attrib);
    gsub(/(^ | $)/, "", attrib);
    if ( match(code, "(^|\n)" guard "+(\n|$)" ) ) {
      len = RLENGTH; st = RSTART;
      return "<pre><code class=\"" attrib "\">" HTML( substr(code, 1, st - 1) ) "</code></pre>\n" \
             _block( substr( code, st + len ) );
    } else {
      match( block, /(^|\n)[[:space:]]*(\n|$)/ ) || match( block, /$/ );
      len = RLENGTH; st = RSTART;
      return "<p>" inline( substr(block, 1, st - 1) ) "</p>\n" \
             _block( substr(block, st + len) );
    }

  # First Order Heading H1 + Attrib
  } else if ( match( block, /^([^\n]+)([ \t]*\{([^\}\n]+)\})\n===+(\n|$)/ ) ) {
    len = RLENGTH; text = attrib = block;
    sub(/([ \t]*\{([^\}\n]+)\})\n===+(\n.*)?$/, "", text);
    sub(/\}\n===+(\n.*)?$/, "", attrib); sub(/^([^\n]+)[ \t]*\{/, "", attrib);
    gsub(/[^a-zA-Z0-9_-]+/, " ", attrib); gsub(/(^ | $)/, "", attrib);

    return headline(1, text, attrib) _block( substr( block, len + 1 ) );

  # First Order Heading H1
  } else if ( match( block, /^([^\n]+)\n===+(\n|$)/ ) ) {
    len = RLENGTH; text = substr(block, 1, len);
    sub(/\n===+(\n.*)?$/, "", text);

    return headline(1, text, 0) _block( substr( block, len + 1 ) );

  # Second Order Heading H2 + Attrib
  } else if ( match( block, /^([^\n]+)([ \t]*\{([^\}\n]+)\})\n---+(\n|$)/ ) ) {
    len = RLENGTH; text = attrib = block;
    sub(/([ \t]*\{([^\}\n]+)\})\n---+(\n.*)?$/, "", text);
    sub(/\}\n---+(\n.*)?$/, "", attrib); sub(/^([^\n]+)[ \t]*\{/, "", attrib);
    gsub(/[^a-zA-Z0-9_-]+/, " ", attrib); gsub(/(^ | $)/, "", attrib);

    return headline(2, text, attrib) _block( substr( block, len + 1) );

  # Second Order Heading H2
  } else if ( match( block, /^([^\n]+)\n---+(\n|$)/ ) ) {
    len = RLENGTH; text = substr(block, 1, len);
    sub(/\n---+(\n.*)?$/, "", text);

    return headline(2, text, 0) _block( substr( block, len + 1) );

  # Nth Order Heading H1 H2 H3 H4 H5 H6 + Attrib
  } else if ( match( block, /^(#{1,6})[ \t]*(([^ \t\n]+|[ \t]+[^ \t\n#]|[ \t]+#+[ \t]*[^ \t\n#])+)[ \t]*#*([ \t]*\{([a-zA-Z \t-]*)\})(\n|$)/ ) ) {
    len = RLENGTH; text = attrib = substr(block, 1, len);
    match(block, /^#{1,6}/); n = RLENGTH;

    sub(/^(#{1,6})[ \t]*/, "", text);   sub(/[ \t]*#*([ \t]*\{([a-zA-Z \t-]*)\})(\n.*)?$/, "", text);
    sub(/^(#{1,6})[ \t]*(([^ \t\n]+|[ \t]+[^ \t\n#]|[ \t]+#+[ \t]*[^ \t\n#])+)[ \t]*#*[ \t]*\{/, "", attrib);
    sub(/\})(\n.*)?$/, "", attrib);
    gsub(/[^a-zA-Z0-9_-]+/, " ", attrib); gsub(/(^ | $)/, "", attrib);

    return headline( n, text, attrib ) _block( substr( block, len + 1) );

  # Nth Order Heading H1 H2 H3 H4 H5 H6
  } else if ( match( block, /^(#{1,6})[ \t]*(([^ \t\n]+|[ \t]+[^ \t\n#]|[ \t]+#+[ \t]*[^ \t\n#])+)[ \t]*#*(\n|$)/ ) ) {
    len = RLENGTH; text = substr(block, 1, len);
    match(block, /^#{1,6}/); n = RLENGTH;
    sub(/^(#{1,6})[ \t]*/, "", text); sub(/[ \t]*#*(\n.*)?$/, "", text);

    return headline( n, text, 0 ) _block( substr( block, len + 1) );

  # block images (wrapped in <figure>)
  } else if ( match(block, "^!" lix "\\([\n\t ]*" lid "([\n\t ]+" lit ")?[\n\t ]*\\)(\\{[a-zA-Z \t-]*\\})?(\n|$)") ) {
    len = RLENGTH; text = href = title = attrib = substr( block, 1, len);

    sub("^!\\[", "", text);
    sub("\\]\\([\n\t ]*" lid "([\n\t ]+" lit ")?[\n\t ]*\\)(\\{[a-zA-Z \t-]*\\})?(\n.*)?$", "", text);

    sub("^!" lix "\\([\n\t ]*", "", href);
    sub("([\n\t ]+" lit ")?[\n\t ]*\\)(\\{[a-zA-Z \t-]*\\})?(\n.*)?$", "", href);

    sub("^!" lix "\\([\n\t ]*" lid, "", title);
    sub("[\n\t ]*\\)(\\{[a-zA-Z \t-]*\\})?(\n.*)?$", "", title);
    sub("^[\n\t ]+", "", title);

    sub("^!" lix "\\([\n\t ]*" lid "([\n\t ]+" lit ")?[\n\t ]*\\)", "", attrib);
    sub("(\n.*)?$", "", attrib);
    sub(/^\{[ \t]*/, "", attrib); sub(/[ \t]*\}$/, "", attrib); gsub(/[ \t]+/, " ", attrib);

    if ( match(href, /^<.*>$/) ) { sub(/^</, "", href); sub(/>$/, "", href); }
         if ( match(title, /^".*"$/) ) { sub(/^"/, "", title); sub(/"$/, "", title); }
    else if ( match(title, /^'.*'$/) ) { sub(/^'/, "", title); sub(/'$/, "", title); }
    else if ( match(title, /^\(.*\)$/) ) { sub(/^\(/, "", title); sub(/\)$/, "", title); }

    gsub(/^[\t ]+$/, "", text); gsub(/\\/, "", href);

    return "<figure data-src=\"" HTML(href, 1) "\"" (attrib?" class=\"" HTML(attrib) "\"":"") ">" \
           "<img src=\"" HTML(href, 1) "\" alt=\"" HTML(text?text:title?title:href) "\"" \
           (attrib?" class=\"" HTML(attrib) "\"":"") ">" \
           (title?"<figcaption>" inline(title) "</figcaption>":"") \
           "</figure>\n\n" \
           _block( substr( block, len + 1) );

  # reference style images (block)
  } else if ( match(line, /^!\[([^]]*)\] ?\[([^]]*)\](\n|$)/ ) ) {
    len = RLENGTH;
    text = gensub(/^!\[([^\n]*)\] ?\[([^\n]*)\](\n.*)?$/, "\\1", 1, block);
      id = gensub(/^!\[([^\n]*)\] ?\[([^\n]*)\](\n.*)?$/, "\\2", 1, block);
    if ( ! id ) id = text;
    if ( rl_href[id] && rl_title[id] ) {
      return "<figure data-src=\"" HTML(rl_href[id], 1) "\">" \
               "<img src=\"" HTML(rl_href[id], 1) "\" alt=\"" HTML(text) "\">" \
               "<figcaption>" inline(rl_title[id]) "</figcaption>" \
             "</figure>\n\n" \
             _block( substr( block, len + 1) );
    } else if ( rl_href[id] ) {
      return "<figure data-src=\"" HTML(rl_href[id], 1) "\">" \
               "<img src=\"" HTML(rl_href[id], 1) "\" alt=\"" HTML(text) "\">" \
             "</figure>\n\n" \
             _block( substr( block, len + 1) );
    } else {
      return "<p>" HTML(substr(block, 1, len)) "</p>\n" _block( substr(block, len + 1) );
    }

  # Macros (standalone <<macro>> calls handled as block, so they are not wrapped in paragraph)
  } else if ( match( block, /^<<(([^>]|>[^>])+)>>(\n|$)/ ) ) {
    len = RLENGTH;
    text = gensub(/^<<(([^>]|>[^>])+)>>(\n.*)?$/, "\\1", 1, block);
    return "<code class=\"macro\">" HTML(text) "</code>" _block(substr(block, len + 1) );

  # Definition list
  } else if (match( block, "^(([ \t]*\n)*[^:\n \t][^\n]+\n" \
                           "([ \t]*\n)* ? ? ?:[ \t][^\n]+(\n|$)" \
                          "(([ \t]*\n)* ? ? ?:[ \t][^\n]+(\n|$)" \
                           "|[^:\n \t][^\n]+(\n|$)" \
                           "|( ? ? ?\t|  +)[^\n]+(\n|$)" \
                           "|([ \t]*\n)+( ? ? ?\t|  +)[^\n]+(\n|$))*)+" \
  )) {
    list = substr( block, 1, RLENGTH); block = substr( block, RLENGTH + 1);
    return "\n<dl>\n" _dlist( list ) "</dl>\n" _block( block );

  # Unordered list types
  } else if ( text = _startlist( block, "ul", "-",   "([+*•]|[0-9]+\\.|#\\.|[0-9]+\\)|#\\))") ) {
    return text;
  } else if ( text = _startlist( block, "ul", "\\+", "([-*•]|[0-9]+\\.|#\\.|[0-9]+\\)|#\\))") ) {
    return text;
  } else if ( text = _startlist( block, "ul", "\\*", "([-+•]|[0-9]+\\.|#\\.|[0-9]+\\)|#\\))") ) {
    return text;
  } else if ( text = _startlist( block, "ul", "•", "([-+*]|[0-9]+\\.|#\\.|[0-9]+\\)|#\\))") ) {
    return text;

  # Ordered list types
  } else if ( text = _startlist( block, "ol", "[0-9]+\\.", "([-+*•]|#\\.|[0-9]+\\)|#\\))") ) {
    return text;
  } else if ( text = _startlist( block, "ol", "[0-9]+\\)", "([-+*•]|[0-9]+\\.|#\\.|#\\))") ) {
    return text;
  } else if ( text = _startlist( block, "ol", "#\\.", "([-+*•]|[0-9]+\\.|[0-9]+\\)|#\\))") ) {
    return text;
  } else if ( text = _startlist( block, "ol", "#\\)", "([-+*•]|[0-9]+\\.|#\\.|[0-9]+\\))") ) {
    return text;

  # Split paragraphs
  } else if ( match( block, /(^|\n)[[:space:]]*(\n|$)/) ) {
    len = RLENGTH; st = RSTART;
    return _block( substr(block, 1, st - 1) ) "\n" \
           _block( substr(block, st + len) );

  # Horizontal rule
  } else if ( match( block, /(^|\n) ? ? ?((\* *){3,}|(- *){3,}|(_ *){3,})($|\n)/) ) {
    len = RLENGTH; st = RSTART;
    return _block(substr(block, 1, st - 1)) "<hr>\n" _block(substr(block, st + len));

  # Plain paragraph
  } else {
    return "<p>" inline(block) "</p>\n";
  }
}

function _startlist(block, type, mark, exclude, LOCAL, st, len, list, indent, text) {
  if (match( block, "(^|\n) ? ? ?" mark "[ \t][^\n]+(\n|$)" \
                                   "(([ \t]*\n)* ? ? ?" mark "[ \t][^\n]+(\n|$)" \
                                   "|([ \t]*\n)*( ? ? ?\t|  +)[^\n]+(\n|$)" \
                                   "|[^\n \t][^\n]+(\n|$))*" ) ) {
    st = RSTART; len = RLENGTH; list = substr( block, st, len);

    sub("^\n", "", list); match(list, "^ ? ? ?"); indent = RLENGTH;
    gsub( "(^|\n) {0," indent "}", "\n", list); sub("^\n", "", list);

    text = substr(block, 1, st - 1); block = substr(block, st + len);
    if (match(text, /\n[[:space:]]*\n/)) return 0;
    if (match(text, "(^|\n) ? ? ?" exclude "[ \t][^\n]+")) return 0;
    if (match( list, "\n" exclude "[ \t]" )) {
      block = substr(list, RSTART + 1) block;
      list = substr(list, 1, RSTART);
    }

    return _block( text ) "<" type ">\n" _list( list, mark ) "</" type ">\n" _block( block );
  } else return 0;
}

function _list (block, mark, p, LOCAL, len, st, text, indent, task) {
  if ( match(block, "^([ \t]*\n)*$")) return;

  match(block, "^" mark "[ \t]"); indent = RLENGTH;
  sub("^" mark "[ \t]", "", block);

  if (match(block, /\n[ \t]*\n/)) p = 1;

  match( block, "\n" mark "[ \t][^\n]+(\n|$)" );
  st = (RLENGTH == -1) ? length(block) + 1 : RSTART;
  text = substr(block, 1, st); block = substr(block, st + 1);

  gsub("\n {0," indent "}", "\n", text);

  task = match( text, /^\[ \]/   ) ? "<li class=\"task pending\"><input type=checkbox disabled>"      : \
         match( text, /^\[-\]/   ) ? "<li class=\"task negative\"><input type=checkbox disabled>"     : \
         match( text, /^\[\/\]/  ) ? "<li class=\"task partial\"><input type=checkbox disabled>"      : \
         match( text, /^\[\?\]/  ) ? "<li class=\"task unsure\"><input type=checkbox disabled>"       : \
         match( text, /^\[[xX]\]/) ? "<li class=\"task done\"><input type=checkbox disabled checked>" : "<li>";
  sub(/^\[[-? \/xX]\]/, "", text);

  text = _nblock( text );
  if ( ! p && match( text, "^<p>(</p[^>]|</[^p]|<[^/]|[^<])*</p>\n$" ))
     gsub( "(^<p>|</p>\n$)", "", text);

  return task text "</li>\n" _list(block, mark, p);
}

function _dlist (block, LOCAL, len, st, text, indent, p) {
  if (match( block, "^([ \t]*\n)*[^:\n \t][^\n]+\n" )) {
    len = RLENGTH; text = substr(block, 1, len);
    gsub( "(^\n*|\n*$)", "", text );
    return "<dt>" inline( text ) "</dt>\n" _dlist( substr(block, len + 1) );
  } else if (match( block, "^([ \t]*\n)* ? ? ?:[ \t][^\n]+(\n|$)" \
                         "([^:\n \t][^\n]+(\n|$)" \
                         "|( ? ? ?\t|  +)[^\n]+(\n|$)" \
                         "|([ \t]*\n)+( ? ? ?\t|  +)[^\n]+(\n|$))*" \
  )) {
    len = RLENGTH; text = substr(block, 1, len);
    sub( "^([ \t]*\n)*", "", text);
    match(text, "^ ? ? ?:(\t| +)"); indent = RLENGTH;
    sub( "^ ? ? ?:(\t| +)", "", text);
    gsub( "(^|\n) {0," indent "}", "\n", text );

    text = _nblock(text);
    if (match( text, "^<p>(</p[^>]|</[^p]|<[^/]|[^<])*</p>\n$" ))
       gsub( "(^<p>|</p>\n$)", "", text);

    return "<dd>" text "</dd>\n" _dlist( substr(block, len + 1) );
  }
}

BEGIN {
  # Global Vars
  file = ""; rl_href[""] = ""; rl_title[""] = "";
  if (ENVIRON["MD_HTML"] == "true") { AllowHTML = "true"; }
  HL[1] = 0; HL[2] = 0; HL[3] = 0; HL[4] = 0; HL[5] = 0; HL[6] = 0;
  # hls = "0 0 0 0 0 0";

  # Universal Patterns
  nu = "(\\\\\\\\|\\\\[^\\\\]|[^\\\\_]|_[[:alnum:]])*"    # not underline (except when escaped)
  na = "(\\\\\\\\|\\\\[^\\\\]|[^\\\\\\*])*"  # not asterisk (except when escaped)
  ieu =  "_([^_[:space:]]|[^_[:space:]]" nu "[^_[:space:]])_"                 # inner <em> (underline)
  isu = "__([^_[:space:]]|[^_[:space:]]" nu "[^_[:space:]])__"                # inner <strong> (underline)
  iea =    "\\*([^\\*[:space:]]|[^\\*[:space:]]" na "[^\\*[:space:]])\\*"     # inner <em> (asterisk)
  isa = "\\*\\*([^\\*[:space:]]|[^\\*[:space:]]" na "[^\\*[:space:]])\\*\\*"  # inner <strong> (asterisk)

  lix="\\[(\\\\[^\n]|[^]\n\\\\[])*\\]"  # link text
  lid="(<(\\\\[^\n]|[^\n<>\\\\])*>|(\\\\.|[^()\"'\\\\])+|([^<\n\t ()\\\\]|\\\\[^\n])(\\\\[\n]|[^\n\t \\(\\)\\\\])*)"  # link dest
  lit="(\"(\\\\.|[^\"\\\\])*\"|'(\\\\.|[^'\\\\])*'|\\((\\\\.|[^\\(\\)\\\\])*\\))"  # link text
  # link text with image def
  lii="\\[(\\\\[^\n]|[^]\n\\\\[])*(!" lix "\\([\n\t ]*" lid "([\n\t ]+" lit ")?[\n\t ]*\\))?(\\\\[^\n]|[^]\n\\\\[])*\\]"

  # Buffering of full file ist necessary, e.g. to find reference links
  while (getline) { file = file $0 "\n"; }
  # Clean up MS-DOS line breaks
  gsub(/\r\n/, "\n", file);

  # Fill array of reference links
  f = file; rl_id;
  re_reflink = "(^|\n) ? ? ?\\[([^]\n]+)\\]: ([^ \t\n]+)(\n?[ \t]+(\"([^\"]+)\"|'([^']+)'|\\(([^)]+)\\)))?(\n|$)";
  # /(^|\n) ? ? ?\[([^]\n]+)\]: ([^ \t\n]+)(\n?[ \t]+("([^"]+)"|'([^']+)'|\(([^)]+)\)))?(\n|$)/
  while ( match(f, re_reflink ) ) {
    rl_id           = gensub( re_reflink, "\\2", 1, substr(f, RSTART, RLENGTH) );
    rl_href[rl_id]  = gensub( re_reflink, "\\3", 1, substr(f, RSTART, RLENGTH) );
    rl_title[rl_id] = gensub( re_reflink, "\\5", 1, substr(f, RSTART, RLENGTH) );
    f = substr(f, RSTART + RLENGTH);
    rl_title[rl_id] = substr( rl_title[rl_id], 2, length(rl_title[rl_id]) - 2 );
    if ( rl_href[rl_id] ~ /<.*>/ ) rl_href[rl_id] = substr( rl_href[rl_id], 2, length(rl_href[rl_id]) - 2 );
  }
  # Clear reflinks from File
  while( gsub(re_reflink, "\n", file ) );
  # for (n in rl_href) { debug(n " | " rl_href[n] " | " rl_title[n] ); }

  # Run Block Processing -> The Actual Markdown!
  printf "%s", _nblock( file );
}
